// 																																			KEDS Project
// 																												(Kansas Event Data System)
//__________________________________________________________________________________
/* 																																		TestRandomizer 

This program randomizes the question and answer ordering in multiple choice exams,
thereby allowing the instructor to spend more time generating and analyzing event 
data.

NOTE: I'm not a big fan of multiple-choice tests and I only use them at the 
      introductory level -- that is, large lecture classes.  Upper division
      classes get objective tests and lots of papers.  However, all of the 
      evidence indicates that multiple-choice tests work fine as a *diagnostic*
      instrument, the students do written work in their discussion sections,
      and by using machine-graded multiple choice exams, I save the graduate 
      students a huge amount of work. 

WHY?
There are plenty of more sophisticated systems available as either shareware or 
commercial software that do the same thing that this program will do (and a whole 
lot more).  I couldnt find one for the Macintosh that allowed me to read and writing 
the questions in ASCII, so I wrote this one. 

KEYBOARD INPUT:
Program prompts for four items:

	"Enter name of file containing test information:" -- Enter the name of the 
			file containing test information, followed by <return>.  This should be
			in the same directory/folder as the program

	"Enter prefix for output files:" -- Enter the string that will be used to
			identify the various output files.
			 
	"Create multiple, randomized exams? (Y/N):" -- Enter "Y" or "N" followed
			by <return>.  If "Y" is entered, NTEST (currently = 4) exams are 
			created with a different, randomized  order of questions and answers.
			This option is used to produce the actual tests.
			If "N" is entered, the questions are listed in the same order as 
			they are found in the input file -- this option is useful when you
			are trying to figure out which questions to put in the test.

	"Create keyed copy of Exam 1? (Y/N):" -- Enter "Y" or "N" followed
			by <return>.  If "Y" is entered, a file will be created with a
			".answer" suffix that contains the first version of the exam with
			"*" before the correct answer.
						
COMMAND LINE ARGUMENTS:
The program accepts optional command line arguments in the form

./TestRandomizer [file_name] [-d] [-f]

file_name:
     The name of the file containing test information. This should be in the 
     same directory/folder as the program.
     
-d
     Draft mode: produces a single copy of the exam and a copy of the correct 
     answer indicated by *.
     
-f
     Final mode: produces four randomized coped of the exam and a copy of the 
     correct answer indicated by *.

SCREEN OUTPUT:
Just enough to show the program is running.  Program issues a note when it 
doesn't recognize a category, and when it is writing test files.  The program 
should run very quickly: the example file runs in less than a second on a 
350Mhz Macintosh G3.

INPUT FILE:
The input is from a single ASCII "text" file; "000.Midterm.testbank" gives
an example of an input file.

The first set of records gives the "categories" for the questions.  These are
strings (up to 15 characters) that classify each question -- for example,
which lecture or text chapter it relates to.  The category set is terminated
by a line beginning with "-".

A set of "comment" lines -- blank lines or lines beginning with a "/" -- can come 
before and after the categories, and blank lines can be used to separate groups of 
categories.

The remainder of the file consists of question blocks.  These have the 
following format:

  category line
  question
  response 1  [default correct answer]
  response 2
  response 3 
  response 4 
  response 5 
    
Question blocks are separated by at least one blank line or comment line; more
than one blank/comment can separate the blocks.

The questions and responses need to be a *single line of text*, with no embedded
<returns> (so-called "hard line breaks") .  Most word processors "word wrap" with
"soft returns" that depend on the display window size; if possible, turn this 
feature off and check to make sure none of the text contains <return>s except at 
the end of the line.

The "category line" should contain one of the categories listed at the beginning
of the file.  It can contain additional information as well; for example

               L-HIST-25  01.02.14

might indicate question #25 for the "Lecture-History" part of the course, and was
entered 14 Feb 2001.  It matches the category "L-HIST".  The matching is done
by looking for the category *anywhere* in the string, so "25-L-HIST" would also
work.  [Also see note below]

Questions are selected by putting a "+" at the beginning of the category line.
For example:

								+++L-HIST-25  01.02.14
								
If there is no + at the beginning of the category line, the question will not
be included in the test.  By using a string that is not found in the questions
-- for example the '+++' used above -- the file can be "reset" by just doing 
a global replace to get rid of these markers (or replace them with an indicator 
showing when they were used -- see note below).

The "question" is simply the stimulus for the question.

The program assumes exactly 5 responses for each question (this can be changed
globally using the constant N_RESPN).  Two modifiers can be used on responses:

# -- this indicates the correct answer.  If no response is tagged with a #, the 
     first response is assumed to be the correct one but an error message
     is issued saying that the question does not have a tagged answer.
     Experience to date has shown that using defaults is NOT a good idea since
     these are easily confused with cases where one forgot to mark the correct
     answer.
     
@ -- response is fixed in place: this can be used for "All of the above" and
     "None of the above" responses.  These must be the final responses in the
     list -- that is, once the system fixes the location of one response, it
     expects all subsequent responses to be fixed.  If a fixed response is
     the correct answer, use "#@" (not "@#").  See partial error checking on 
     this below.
     
The question can have an optional single-character designator followed by a 
period; this is skipped over.

Example of a question block:

++D-W4-10 
Walter B. Wriston argues that the technological revolution:
a.  Will change the manner in which we conduct business, but not much else
#b.  Means that information will become our greatest source of wealth
c.  Will eventually weaken the power of the individual
@d.  All of the above
@e.  None of the above

OUTPUT FILES:
report file : This gives a listing of the number of questions in each 
               category.  It has the name of the output file plus the 
               suffix ".report".

test file    : This gives the randomized test. The suffix is ".exam.N"
							 where N is 1, 2, 3, 4.

key file     : This gives the key for the test. The suffix is ".key.N"

answer file  : This gives Exam 1 with a '*' in front of the correct
               answer. It has the name of the output file plus the 
               suffix ".answer".
							
The program is currently set to generate 4 different tests.  The random number 
generator is reset at the beginning of the program, so additional tests could be 
created by re-running the program.

ADDITIONAL NOTES:

1. A "blank line" is a line containing only a <return>.  Blank lines 
   in the category list are internally represented with _-* (BLANK_TOKEN), 
   so don't use this string as a category label.
   
   [As a bit of error-checking, the program will in fact treat as a "blank line"
   any line containing only spaces and tabs.  Such lines *appear* blank in
   most word processors and consequently are difficult to track down.] 

2. When figuring out the category labels, remember that a category is 
   matched anywhere in the string, so a category label R-PR12 would be 
   matched by R-PR1.  However, the matches are tested in the order they
   are entered, so if R-PR12 came before R-PR1, the counts would be 
   correct.
   
3. To reset a file with labels, use a grep-compatible search-and-replace
   with something like
   
   Find:     \++(.+)
   Replace:  \1 MT2.05
   
   This removes the selection indication (assuming it consists of 
   consecutive +'s), then adds a string indicating when the question was used,
   e.g. "MT2.05" = second midterm, 2005
   
INTERFACING WITH MS-WORD

In order to produce a nice-looking exam, I use two formats: 'qresp' for the
responses and 'quest' for the questions. After copying the questions into
a Word document, I first set everything to the 'qresp' style, then do a
global replace of 

    Find: ^#.        (this is Word's equivalent to a grep \d\.)
    Replace: Style quest

One still has to go through and manually get the page-breaks correct, but
otherwise this is fairly quick and reliable.

PROGRAM LIMITATIONS
maximum length of input string			 					 1023
maximum number of characters in exam					32768  
maximum length of a category string							 15
maximum number of question categories						 64 
maximum number of questions											256 

ERROR CHECKING

1. If an answer contains "All of " or "None of " (as in "All of the above")
   and the fixed-response operator @ has not been used on any of the responses,
   the summary file will contain an alert to this effect.  This doesn't
   catch all possible problems, but gets the worse of them.

2. If no correct answer is marked with #, an alert is issued, and the first
   response is considered the default correct answer. Prior to introducing
   this alert, the default answer option has been the source of several 
   embarrassing errors where, in fact, the first answer was not correct,
   but no correct answer had been marked.
   
3. Beyond this, there is very little error checking.  The system will terminate 
   when it can't open files, and notes when there are unrecognized categories,
   but otherwise does no checking for internal consistency and correct formatting.  
   Usually formatting errors (e.g. unwanted <returns> in the text) will show up 
   as sentence fragments -- which, unfortunately, may even look plausible --
   rather than crashing the program.

PROGRAM LOGIC

1. The only vaguely subtle aspect of the program is that all of the text
   is stored as null-terminated strings in one big block of memory (astchar[]),
   then accessed via pointers (actually, array indices) into this.  Same
   scheme as used in KEDS and TABARI (surprise...)
   
2. The array aquest[] is a structure that holds all of the information
   relevant to the questions.
   
3. The same "shuffling" algorithm is used in ShuffleQuestion and ShuffleAnswer;
   there may be some more efficient way to do this but it is good enough.
   
4. Program run time is dominated by input and output operations, and 
   therefore I didn't do much to optimize the string manipulation (which 
   is minimal in any case). 
   
5. This program occasionally has been affecting the Mac OS9.1 operating system
   as if it had a memory leak (or bounds error) somewhere.  If you happen to
   find this, please let me know.


SYSTEM REQUIREMENTS:
This program was originally developed for the Macintosh operating system (OS 9)
but contains no Macintosh-specific features and will presumably work on any
system that has a C compiler.  Version 0.7 is running under OS-X

Compiler:		Metrowerks C (original)
            gcc (version 0.7)
Language:		ANSI C with C++ comments
For plausible indenting of this source code, set the tab size in your editor to "2"

REVISION HISTORY
  19-March-01:  Initial version (0.5B1)
   5-April-02:  Non-randomized output (0.5B2)
  30-March-03:  .answer file; "all/none" checking (0.6B1)
27-October-05:  unused options eliminated; unmarked answer alert (0.7B1)
   4-April-07: Command-line options (0.8B1)
    15-May-07: Bug in draft mode that could duplicate some answers (0.8B2)

PROVENANCE
Programmer: Philip A. Schrodt
            Kansas Event Data System Project
            Center for International Political Analysis
            University of Kansas 
            Lawrence, Kansas, 66044 U.S.A.
            http://www.ku.edu/~keds
						email: schrodt@ku.edu
						
Copyright (c) 2007  Philip A. Schrodt.  All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted under the terms of the GNU General Public License:
http://www.opensource.org/licenses/gpl-license.html

Report bugs to: schrodt@ku.edu

*/

//__________________________________________________________________________________

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

//__________________________________________________________________________________
// 	Macros

#define	FALSE			0
#define TRUE			1

//#define DO_HTML       // option for HTML coding

#define IN_FILE			"000.Midterm.testbank" // Debugging
#define OUT_FILE		"0.randomizertest" 
#define TEST_TITLE	"Random Test"

#define SUMM_SUFFIX	".report"						// suffix for the report file
#define TEST_SUFFIX	".exam.N"  					// suffix for exam files
#define KEY_SUFFIX	".key.N"  					// suffix for key files
#define BLANK_TOKEN "_-*"								// token to indicate blank lines in category list		

#define	MAX_INPUT				 1024     // maximum length (plus one) of input string
#define	MAX_STORE				32768     // maximum number of characters in exam
#define	MAX_CAT_LENGTH		 16     // maximum length of a category string (plus 1)
#define	MAX_CATEGORY		   64     // maximum number of question categories
#define	MAX_QUESTN		    256     // maximum number of questions
#define MAX_FILE_NAME      32     // maximum length of file name

#define N_TESTS							4			// number of tests produced
#define N_RESPN							5			// number of responses per question

//___________________________________________________________________________
// Prototypes

void GetFileNames();  				// get names of files from keyboard
void Report_Error(char s[]);  // exit on early end-of-file
void SkipComments(void);    	// skip comment and blank lines
void ReadInputFile(void);   	// read input file
void WriteSummary(void);  		// writes the category totals
void ShuffleQuestions(void); 	// randomly orders the questions
void ShuffleAnswers(int kquest); 	// randomly orders the answers of question kquest
void WriteTest(void);					// writes a test in randomized order
void WriteTestHTML(void);			// writes a test with HTML tags

//__________________________________________________________
// Globals


FILE	*fin, *ftest, *fkey, *fansw;				// input and output files
char  sfin_name[MAX_FILE_NAME], sfout_name[MAX_FILE_NAME]; // names of same
char  strin[MAX_INPUT];   				// input string
char  squest[MAX_INPUT];   				// question ID string
int		gotInput = FALSE;						// input name passed via arguments
int		isDraft  = FALSE;						// -d option in arguments
int		isFinal  = FALSE;						// -f option in arguments
int		isHTML = FALSE;							// code styles as HTML
int		isRandom = FALSE;						// produce multiple, randomized tests
int		isAnswer = TRUE;						// produce keyed version of exam 1
int		isProblem = FALSE;					// possible problems with "All of" or "None of" answers

char astchar[MAX_STORE];  // storage for text
int	 nchar = 0;
char acategory[MAX_CATEGORY][MAX_CAT_LENGTH];  // category labels
int  akcategory[MAX_CATEGORY];    // category counts
int  ncat = 0;										// number of categories found

struct {
	int istart;    				// start of question text
	int iresp[N_RESPN];  	// start of each response
	int icorrect;					// correct response (defaults to 0)
	int nmove;     				// number of questions that can be moved
} aquest[MAX_QUESTN];
int nquest = 0;

int aorder[MAX_QUESTN]; // random ordering of question
int answer[N_RESPN];  	// random ordering of answers
	
//__________________________________________________________________________________
//																																			main program

int main(int argc, char *argv[])
{
int 	ka = 1;
char	sfkey_name[32], sftest_name[32], sfansw_name[32];
time_t thetime;

srand(time(NULL));												// randomly seed the RNG

while(--argc) {     // process the arguments
	if ('-' != argv[ka][0]) {
		strcpy(sfin_name,argv[ka]);
		printf("Input file is\"%s\"\n",sfin_name);
		gotInput = TRUE;
	}
	else {
		if ('d' == argv[ka][1]) { 
			isDraft = TRUE;
			printf("Draft production mode\n");
		}
		if ('f' == argv[ka][1]) { 
			isFinal = TRUE;
			printf("Final production mode\n");
		}
	}
  ++ka;
}

GetFileNames();

strcpy(sftest_name,sfout_name);
strcat(sftest_name,SUMM_SUFFIX);   // open the report file
ftest = fopen(sftest_name,"w");
if (!ftest) { printf("Error: can't open %s",sftest_name); exit(0); }
time(&thetime);
fprintf(ftest, "TestRandomizer Report\nDate:%s\nInput File:%s\n",ctime(&thetime),sfin_name);
fprintf(ftest, "\nQuestions in exam:\n");

ReadInputFile();
WriteSummary();

fclose(ftest);

strcpy(sftest_name,sfout_name);			// copy of file prefix
strcat(sftest_name,TEST_SUFFIX);	// add suffix
strcpy(sfkey_name,sfout_name);
strcat(sfkey_name,KEY_SUFFIX);

for (ka = 1; ka <= N_TESTS; ++ ka) {
	sftest_name[strlen(sftest_name)-1] = ka + 48; //  set numerical suffix; open the test file
	ftest = fopen(sftest_name,"w");
	if (!ftest) { printf("Error: can't open %s",sftest_name); exit(0); }
	printf("Writing %s\n",sftest_name);
	fprintf(ftest,"Test Version %d\n\n",ka);

	sfkey_name[strlen(sfkey_name)-1] = ka + 48;  // same for the key file
	fkey = fopen(sfkey_name,"w");
	if (!fkey) { printf("Error: can't open %s",sfkey_name); exit(0); }
	fprintf(fkey,"Key for Test Version %d\n\n",ka);
	
	if (isAnswer) {
		strcpy(sfansw_name,sfout_name);
		strcat(sfansw_name,".answers");		
		fansw = fopen(sfansw_name,"w");
		if (!fansw) { printf("Error: can't open %s",sfansw_name); exit(0); }
		printf("Writing %s\n",sfansw_name);
		fprintf(fansw,"Answers to Test Version %d\n\n",ka);
	}

	ShuffleQuestions();
	if (isHTML) WriteTestHTML();
	else  WriteTest();
	
	fclose(ftest);
	fclose(fkey);
	if (isAnswer) {
		fclose(fansw);
		isAnswer = FALSE;
	}
	
	if (!isRandom) break;  // only produce one version
}	
printf("\n");
	
} // main

void GetFileNames()  
// get names of files from keyboard
{
int ka;
	printf("TestRandomizer:  Program for generating multiple choice tests with \n");
	printf("Version 0.8B1    randomized question and response ordering\n");
	printf("                 Kansas Event Data System (KEDS)\n");
	printf("\nCopyright 2007, Philip A. Schrodt\n");
	printf("Redistribution and use in source and binary forms, with or without modification,\n");
	printf("are permitted under the terms of the GNU General Public License:\n");
	printf("http://www.opensource.org/licenses/gpl-license.html\n\n");

#if TRUE					// get names from keyboard
	if (!gotInput) {
		printf("\aEnter name of file containing test information:");
		gets(sfin_name);
	}
	printf("Enter prefix for output files:");
  gets(sfout_name);

	if (isDraft) isAnswer = TRUE;  // this is actually just the default' does nothing
	else if (isFinal) isRandom = TRUE;
	else {
		printf("Create multiple, randomized exams? (Y/N):");
		gets(strin);
		if (('n' == strin[0]) || ('N' == strin[0])) isRandom = FALSE;
		else isRandom = TRUE;
	
		printf("Create keyed copy of Exam 1? (Y/N):");
		gets(strin);
		if (('n' == strin[0]) || ('N' == strin[0])) isAnswer = FALSE;
		else isAnswer = TRUE;
	}

#ifdef DO_HTML
	printf("Include HTML formatting in output? (Y/N):");
	gets(strin);
	if (('n' == strin[0]) || ('N' == strin[0])) isHTML = FALSE;
	else isHTML = TRUE;
#endif
	
#else							// get names from macros (debugging)
	strcpy(sfin_name,IN_FILE);
	strcpy(sfout_name,OUT_FILE);
	isHTML = FALSE;
#endif
} // GetFileNames
	
void Report_Error(char s[])  
// exit on early end-of-file or other fatal errors
{
	printf("\a%s\nLast line read was\n%s\n",s,strin);
	printf("Press <return> to exit");
	gets(strin);
	exit(0);
} // Error_Exit
	
void SkipComments(void)    
// skip comment and blank lines; leaves a non-blank line or eof in strin
{
int ka = 0;

	do {
		fgets(strin,MAX_INPUT,fin);
//		printf("Skip:%s",strin);
		ka = 0;
		while ((' ' == strin[ka]) || ('\t' == strin[ka])) ++ka;   // skip initial white space
	} 
	while (!feof(fin)            && 
	        (('\n' == strin[ka]) ||   // blank line
	         ('/' == strin[ka]))      // comment line
	      ); 
	  
} // SkipComments

void ReadInputFile(void)   
// read input file
// Note: this should check for overflow in astchar
{
int ka;
int ks;
char * ps;
int hasNoMove = FALSE;   // answers have no-move operators
int hasAllNone = FALSE;  // answer containing "All of " or "None of " in an answer


	fin = fopen(sfin_name,"r");
	if (!fin) { printf("Error: can't open %s",sfin_name); exit(0); }
	for (ka = 0; ka < MAX_CATEGORY; ++ka) akcategory[ka] = 0;
	
	SkipComments();    // skip opening comments
	
	while (!feof(fin) && ('-' != strin[0]) && (ncat < MAX_CATEGORY)) {	   // read categories list
	  if ('\n' != strin[0]) {
			strin[strlen(strin)-1] = '\0';   // get rid of the \n
			strcpy(acategory[ncat],strin);
		} 
		else strcpy(acategory[ncat],"BLANK_TOKEN");		
		++ncat;
		fgets(strin,MAX_INPUT,fin);
	}
	if (ncat >= MAX_CATEGORY) {
		printf("\aMaximum number of categories exceeded; remaining categories skipped\n");
		while (!feof(fin) && ('-' != strin[0]))		fgets(strin,MAX_INPUT,fin);
	} 
	if (feof(fin)) Report_Error("Unexpected end of file reading categories");
	strcpy(acategory[ncat],"Unrecognized");
		
	SkipComments(); 
	while (!feof(fin)) { // main question-reading loop
		if ('+' == strin[0]) {  // select question
			ps = strin;
			while ('+' == *ps) ++ps;   // skip over +'s
			strcpy(squest,ps);
			fprintf(ftest, "%s",squest);
			ka = 0;
			while ((ka < ncat)  && !strstr(ps,acategory[ka])) ++ka;  // find the category
			if (ka < ncat ) ++akcategory[ka];
			else {
			  printf("Unrecognized category %s\n",strin);
			  ++akcategory[ncat];
			}
			
			fgets(strin,MAX_INPUT,fin);  // read question
			aquest[nquest].istart = nchar;
			ks = 0;
			while (strin[ks]) astchar[nchar++] = strin[ks++];
			astchar[nchar-1] = '\0';   // replace \n with \0
			if (nchar >= MAX_STORE) Report_Error("Out of memory for storing exam text");
			// read and process the responses
			aquest[nquest].icorrect = -1;	// indicator that correct response has not been entered
			aquest[nquest].nmove = 0;

			for (ka = 0; ka < N_RESPN; ++ka) {
				fgets(strin,MAX_INPUT,fin);  // read response
				ks = 0;
				if ('#' == strin[0]) {   // tag for correct answer
				  ++ks;
					aquest[nquest].icorrect = ka;	// record correct response
				}

				if ('@' == strin[ks]) {
					++ks;  // non-moving answer
					hasNoMove  = TRUE;
				}				 
				else ++aquest[nquest].nmove;
				
				while ((' ' == strin[ks]) || ('\t' == strin[ks])) ++ks;  // skip white space following indicator
				
				if ('.' == strin[ks+1]) ks += 2; // skip designator, e.g. "a.", "b." etc.
				else {     // diagnostic for possible multiple lines due to formatting errors
					printf("Possible multiple line question or response in question %d: %s\n",(nquest+1),squest);
					fprintf(ftest, "Possible multiple line question or response in question %d: %s\n",(nquest+1),squest);
					isProblem = TRUE;
				}
				while (' ' == strin[ks]) ++ks;  // skip remaining leading blanks
				   
				if (!hasNoMove) {  // check for "All/None of the above" missing no-move operator
					if (strstr(&strin[ks],"All of ")) hasAllNone = TRUE;
					else if (strstr(&strin[ks],"None of ")) hasAllNone = TRUE;
					else if (&strin[ks] == strstr(&strin[ks],"all of ")) hasAllNone = TRUE;
					else if (&strin[ks] == strstr(&strin[ks],"none of ")) hasAllNone = TRUE;
				}
				aquest[nquest].iresp[ka] = nchar;
				while (strin[ks])  astchar[nchar++] = strin[ks++];  // transfer response
				astchar[nchar-1] = '\0';   // replace \n with \0
			}
			if (hasAllNone) {
			  printf("Possible \"All of\"/\"None of\" problem in question %d: %s\n",(nquest+1),squest);
			  fprintf(ftest, "Possible \"All of\"/\"None of\" problem in question %d: %s\n",(nquest+1),squest);
			  isProblem = TRUE;			  
			}
			if (aquest[nquest].icorrect < 0) {
			  printf("Correct answer has not been tagged in question %d: %s\n",(nquest+1),squest);
			  fprintf(ftest, "Correct answer has not been tagged in question %d: %s\n",(nquest+1),squest);
			  aquest[nquest].icorrect = 0;  // use a default to keep the program running
			  isProblem = TRUE;			  
			}
			hasNoMove  = FALSE; 
			hasAllNone = FALSE;
			++nquest;
		} 
		else {  // skip this question or to next blank line
			for (ka = 0; ka < N_RESPN+1; ++ka) {
				fgets(strin,MAX_INPUT,fin);
				if ('\n' == strin[0]) break;
			}
		}
		SkipComments(); 
	} // while

	fclose(fin);
	
} // ReadInputFile

void WriteSummary(void)  
// write number of questions by category
{
int ka;

	fprintf(ftest,"\nNumber of questions by category\n");
	ka = 0;
	while (ka < ncat) {
	  if (strcmp("BLANK_TOKEN",acategory[ka])) {  
	  	fprintf(ftest,"%-16s\t%3d\n",acategory[ka],akcategory[ka]);
	  	printf("%-16s   %3d\n",acategory[ka],akcategory[ka]);
	  }
	  else { // display a blank line
	  	fprintf(ftest,"\n");
	  	printf("\n");
	  }
	  ++ka;
	}	
	fprintf(ftest,"\n%-16s\t%3d\n",acategory[ncat],akcategory[ncat]);  // display "Unrecognized" count
	printf("\n%-16s   %3d\n",acategory[ncat],akcategory[ncat]);

	fprintf(ftest,"\nTotal           \t%3d\n",nquest);  // display total
	printf("\nTotal             \t%3d\n\n",nquest);
						
	if (isProblem) {
		fprintf(ftest, "\nPossible \"All of\"/\"None of\", multiple line or unmarked answer problems detected!\n\n");
		printf("Possible \"All of\"/\"None of\", multiple line or unmarked answer problems detected!\n\n");
  }
	
	fprintf(ftest,"\nMemory used for exam \t%6d characters\nMemory remaining     \t%6d characters\n",nchar,MAX_STORE-nchar);  // display total
	printf("\nMemory used for exam  %6d chars\nMemory remaining      %6d chars\n",nchar,MAX_STORE-nchar);  // display total
} // WriteSummary
	
void ShuffleQuestions(void)
// randomly orders the questions; returns this in aorder[]
{
int aused[MAX_QUESTN];  // has question been used?
int ka, kb;
int notused = nquest;
int irand;

	if (!isRandom) {  // do not randomize -- list questions as found
		for (ka = 0; ka < nquest; ka++) aorder[ka] = ka;
		return;		
	}
	
	for (ka = 0; ka < nquest; ka++) aused[ka] = 0;

	while (notused) {
		irand = (rand() % notused) + 1;  // get random integer between 1 and notused
		ka = 0;		// find the irand-th unused question
		kb = 0;
		while (kb < irand) { if (!aused[ka++]) ++kb;}
		aorder[--notused] = ka - 1;
		aused[ka-1] = 1;
	}

} // ShuffleQuestions

void ShuffleAnswers(int kquest)
// randomly orders the answers of question kquest; 
// returns this in answer[]
{
int aused[N_RESPN];  // has question been used?
int ka, kb;
int nm = aquest[kquest].nmove;
int notused = aquest[kquest].nmove;
int irand;

	if (!isRandom) {  // return original order
		for (ka = 0; ka < N_RESPN; ka++) answer[ka] = ka;
		return;		
	}
	
	for (ka = 0; ka < nm; ka++) aused[ka] = 0;

	while (notused) {
		irand = (rand() % notused) + 1;  // get random integer between 1 and notused
		ka = 0;		// find the irand-th unused question
		kb = 0;
		while (kb < irand) { if (!aused[ka++]) ++kb;}
		answer[--notused] = ka - 1;
		aused[ka-1] = 1;
	}

	for (ka = aquest[kquest].nmove; ka < N_RESPN ; ka++) answer[ka] = ka;  // fill in non-moving answers
} // ShuffleAnswers

void WriteTest(void)  
// writes a test in randomized order
{
int ka, kb, kq, kr;

	for (ka = 0; ka < nquest; ka++) {
		kq = aorder[ka];
		fprintf(ftest,"%d. %s\n",ka+1, &astchar[aquest[kq].istart]);
		if (isAnswer) fprintf(fansw,"%d. %s\n",ka+1, &astchar[aquest[kq].istart]);
		fprintf(fkey,"%d. ",ka+1);
		ShuffleAnswers(kq);
		for (kb = 0; kb < N_RESPN; ++kb) {
			kr = answer[kb];
		  fprintf(ftest,"%c. %s\n", (kb+97), &astchar[aquest[kq].iresp[kr]]);
		  if (kr == aquest[kq].icorrect) fprintf(fkey,"%c\n",(kb+97));
			if (isAnswer) {
			  if (kr == aquest[kq].icorrect) fprintf(fansw,"*");
			  fprintf(fansw,"%c. %s\n", (kb+97), &astchar[aquest[kq].iresp[kr]]);
		  }
		}
		fprintf(ftest,"\n");
		if (isAnswer) fprintf(fansw,"\n");
	}
} // WriteTest

void WriteTestHTML(void)  
// writes a test in randomized order with HTML tags
{
int ka, kb, kq, kr;

	for (ka = 0; ka < nquest; ka++) {
		kq = aorder[ka];
		fprintf(ftest,"<H3>%d. %s</H3>\n",ka+1, &astchar[aquest[kq].istart]);
		if (isAnswer)fprintf(fansw,"<H3>%d. %s</H3>\n",ka+1, &astchar[aquest[kq].istart]);
		fprintf(fkey,"%d. ",ka+1);
		ShuffleAnswers(kq);
		for (kb = 0; kb < N_RESPN; ++kb) {
			kr = answer[kb];
		  fprintf(ftest,"<H4>%c. %s</H4>\n", (kb+97), &astchar[aquest[kq].iresp[kr]]);
		  if (kr == aquest[kq].icorrect) fprintf(fkey,"%c\n",(kb+97));
			if (isAnswer) {
			  if (kr == aquest[kq].icorrect) fprintf(fansw,"<H4>*%c. %s</H4>\n", (kb+97), &astchar[aquest[kq].iresp[kr]]);
		    else fprintf(fansw,"<H4>%c. %s</H4>\n", (kb+97), &astchar[aquest[kq].iresp[kr]]);
		  }
		}
		fprintf(ftest,"\n");
	}
} // WriteTestHTML
